import logging
import sys
from os.path import expanduser


class Base(object):

    def __init__(self, **kwargs):
        self._setup_logging(self, **kwargs)

    @staticmethod
    def _setup_logging(self, log_to_screen=False, **kwargs):
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)
        # log setup
        log_location = kwargs.pop("log_path", '/'.join((expanduser('~'), 'logs/pyacm', 'general.log')))
        logger_name = kwargs.pop("name", "base_logger")

        formatter = logging.Formatter(fmt='%(asctime)s %(levelname)-8s %(message)s',
                                      datefmt='%Y-%m-%d %H:%M:%S')
        handler = logging.FileHandler(log_location, mode='w')
        handler.setFormatter(formatter)

        logger = logging.getLogger(logger_name)
        logger.setLevel(logging.INFO)
        logger.addHandler(handler)
        if log_to_screen:
            screen_handler = logging.StreamHandler(stream=sys.stdout)
            screen_handler.setFormatter(formatter)
            logger.addHandler(screen_handler)

        # logging.basicConfig(filename=log_location, level=logging.INFO)
        self.logger = logger


    def info(self, msg):
        self.logger.info(msg)

    def error(self, msg):
        self.logger.error(msg)



# local dynamodb endpoint url
TEST_URL = "http://localhost:8000"

# dynamodb service string literal
DYNAMODB = 'dynamodb'

# source table literal
SOURCE_TABLE = 'source'

# producer config literal
PRODUCER = 'producer'

# consumer config literal
CONSUMER = 'consumer'

# dynamodb last evaluated key literal
LAST_EVALUATED_KEY = 'LastEvaluatedKey'

# thread count literal
THREAD_COUNT = 'TotalSegments'

# scan batch literal
SCAN_BATCH_SIZE = 'Limit'

# function literal
FUNCTION = 'function'

# timeout literal
TIMEOUT = 'timeout'

# args literal
ARGS = 'args'

# index name literal
INDEXNAME = 'IndexName'



class AsyncCounter:
    """Plain async, thread-safe incrementing counter"""
    def __init__(self, initial=0):
        self.value = initial

    def increment(self, num=1):
        self.value += num
        return self.value



import asyncio
from pprint import pprint
import time
import aioboto3
from boto3.dynamodb.conditions import Key

from dynamodb_migration import ddb_const as cst
from dynamodb_migration.aws_base import AWSBase
from dynamodb_migration.counter import AsyncCounter


async def table_query_pk(table, **query):
    if _table := table:
        if pk := query.pop('pk', None):
            if pkv := query.pop('pkv', None):
                return await _table.query(KeyConditionExpression=Key(pk).eq(pkv))
    raise ValueError


async def table_query(table, **query):
    if _table := table:
        return await _table.query(**query)
    raise ValueError


async def index_query_pk(table, **query):
    if _table := table:
        if index_name := query.pop('index_name', None):
            if gsik := query.pop('gsik', None):
                if gsikv := query.pop('gsikv', None):
                    return await _table.query(
                        IndexName=index_name,
                        KeyConditionExpression=Key(gsik).eq(gsikv)
                    )
    raise ValueError


async def index_query(table, **query):
    if _table := table:
        return await _table.query(**query)
    raise ValueError


async def cardinality(response, *consumer_args):
    for _ in response['Items']:
        pprint("current total = {}".format(str(consumer_args[0].increment())))


async def request(segment, exclusive_key, **config):
    if source := config.pop(cst.SOURCE_TABLE, None):
        if exclusive_key:
            return await source.scan(
                Segment=segment,
                ExclusiveStartKey=exclusive_key,
                **config)
        else:
            return await source.scan(
                Segment=segment,
                **config
            )


def init_table(client, **schema):
    try:
        tables = client.list_tables()
        if schema['TableName'] in tables['TableNames']:
            client.delete_table(TableName=schema['TableName'])
            time.sleep(2)
            while True:
                tables = client.list_tables()  # to change
                if schema['TableName'] in tables['TableNames']:
                    time.sleep(3)
                else:
                    break
        client.create_table(**schema)
        waiter = client.get_waiter('table_exists')
        waiter.wait(TableName=schema['TableName'])
        return 0
    except:
        return 1


class DynamoDBClient(AWSBase):
    def __init__(self, queue, local=True, **kwargs):
        super().__init__(**kwargs)
        self.local = local
        self.queue = queue

    def get_dynamodb(self):
        if self.local:
            return aioboto3.resource(cst.DYNAMODB, endpoint_url=cst.TEST_URL)
        else:
            return aioboto3.resource(cst.DYNAMODB)

    def get_client(self):
        if self.local:
            return aioboto3.client(cst.DYNAMODB, endpoint_url=cst.TEST_URL)
        else:
            return aioboto3.client(cst.DYNAMODB)

    async def traverse(self, **config):
        ps = [asyncio.create_task(self.produce(i, **config.get(cst.PRODUCER))) for i in range(config.get(cst.PRODUCER)[cst.THREAD_COUNT])] + [
            asyncio.create_task(self.consume(i, **config.get(cst.CONSUMER))) for i in range(config.get(cst.CONSUMER)[cst.THREAD_COUNT])]

        await asyncio.gather(*ps)

        self.logger.info("Traverse complete...")

    async def produce(self, segment, **config):
        _counter = iteration = 0
        cur = None
        while True:
            response = await request(segment, cur, **config)
            await self.queue.put(response)
            iteration += 1
            _counter += response['Count']

            self.logger.info(f"Producer #{segment} fetches {len(response['Items'])}"
                             f" at iteration {iteration}, total count = {_counter}"
                             f" q size ~ {self.queue.qsize()}")

            if cst.LAST_EVALUATED_KEY not in response:
                break
            else:
                cur = response[cst.LAST_EVALUATED_KEY]
        self.logger.info(f"Producer #{segment} fetched {_counter} items in total.")

    async def consume(self, segment, **config):
        while True:
            try:
                queued_item = await asyncio.wait_for(self.queue.get(), timeout=config[cst.TIMEOUT])
                if queued_item is None:
                    break
            except asyncio.TimeoutError:
                self.logger.error(f"consumer #{segment} timeout...")
                break
            try:
                await config[cst.FUNCTION](queued_item, *config[cst.ARGS])
            except Exception as e:
                self.logger.error(str(e))
                continue
            finally:
                self.queue.task_done()




from datetime import datetime


def profile(func):
    def wrapper(*args, **kwargs):
        import time
        start = time.time()
        func(*args, **kwargs)
        end = time.time()
        print(end - start)
    return wrapper


def time_in_millis(dt):
    return int((dt - datetime.utcfromtimestamp(0)).total_seconds() * 1000)


def now_in_milli():
    return time_in_millis(datetime.utcnow())



import collections
import configparser
import json
from datetime import datetime
from os.path import expanduser

import boto3

from dynamodb_migration.perf import time_in_millis
from dynamodb_migration.base import Base


class AWSBase(Base):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.my_aws_config = configparser.ConfigParser()
        self.my_aws_path = kwargs.pop("my_aws_path", "/".join([expanduser('~'), '.aws/credentials']))
        self.my_aws_config.read(self.my_aws_path)

        profile = kwargs.pop("profile", "default")
        self.my_aws_access_key_id = self.my_aws_config[profile]['aws_access_key_id']
        self.my_aws_secret_access_key = self.my_aws_config[profile]['aws_secret_access_key']

    def get_resource(self, resource_name, **kwargs):
        region = kwargs.pop("region", 'us-east-1')
        mysession = boto3.session.Session(
            aws_access_key_id=self.my_aws_access_key_id,
            aws_secret_access_key=self.my_aws_secret_access_key,
            region_name=region)
        return mysession.resource(resource_name, **kwargs)

    def connect(self, **kwargs):
        region = kwargs.pop("region", 'us-east-1')
        mysession = boto3.session.Session(
            aws_access_key_id=self.my_aws_access_key_id,
            aws_secret_access_key=self.my_aws_secret_access_key,
            region_name=region)
        return mysession.client(**kwargs)

    def connect_with_sts(self, **kwargs):
        aws_identity_path = kwargs.pop("aws_identity_path", "/".join((expanduser('~'), 'aws.identity.json')))

        def _mfa_number():
            try:
                with open(aws_identity_path, 'r') as f:
                    _d = json.load(f)
                return _d['Arn'].replace('user', 'mfa')
            except:
                raise Exception("aws identity not configured")

        aws_path = kwargs.pop("aws_token_path", "/".join((expanduser('~'), 'token.json')))

        with open(aws_path) as f:
            aws_config = json.load(f)

        if not aws_config or 'Credentials' not in aws_config or time_in_millis(
                datetime.strptime(aws_config['Credentials']['Expiration'], '%Y-%m-%d %H:%M:%S+00:00')) <= datetime.now().timestamp() * 1000:
            sts_client_session = boto3.session.Session(
                aws_access_key_id=self.my_aws_access_key_id,
                aws_secret_access_key=self.my_aws_secret_access_key)

            sts_client = sts_client_session.client('sts')

            # default accept code from stdin
            code = input("two factor auth code: ")

            resp = sts_client.get_session_token(
                DurationSeconds=129600,
                SerialNumber=_mfa_number(),
                TokenCode=code
            )
            aws_config = collections.defaultdict(dict)
            aws_config['Credentials']['SessionToken'] = resp['Credentials']['SessionToken']
            aws_config['Credentials']['AccessKeyId'] = resp['Credentials']['AccessKeyId']
            aws_config['Credentials']['SecretAccessKey'] = resp['Credentials']['SecretAccessKey']
            aws_config['Credentials']['Expiration'] = str(resp['Credentials']['Expiration'])

            with open(aws_path, 'w', encoding='utf-8') as f:
                json.dump(aws_config, f)

        region = kwargs.pop("region", 'us-east-1')
        mysession = boto3.session.Session(
            aws_session_token=aws_config['Credentials']['SessionToken'],
            aws_access_key_id=aws_config['Credentials']['AccessKeyId'],
            aws_secret_access_key=aws_config['Credentials']['SecretAccessKey'],
            region_name=region)

        return mysession.client(kwargs.pop('service', 's3'))



import setuptools

setuptools.setup(
    name='dynamodb_traverse',
    version='0.1.1',
    description='High performance, thread safe traversing tool for AWS DynamoDB',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    # The project's main homepage.
    url='https://github.com/holyshipt/dynamodb_traverse',
    # Author details
    author='Lawrence He',
    author_email='ruyangmao1001@gmail.com',
    # Choose your license
    license='MIT',
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.8',

    ],
    packages=setuptools.find_packages(),
    python_requires='>=3.8',
)
